import difflib
import pathlib
import logging
import functools
import subprocess
import panflute as pf

import pandoc_filter

def _check_file_path(file_path:str)->pathlib.Path:
    file_path:pathlib.Path = pathlib.Path(file_path)
    assert file_path.exists()
    assert file_path.is_file()
    return file_path

def _check_the_same_content(file1_path, file2_path):
    with open(file1_path, 'r', encoding='utf-8') as file1:
        content1 = file1.readlines()
    with open(file2_path, 'r', encoding='utf-8') as file2:
        content2 = file2.readlines()
    differ = difflib.Differ()
    diff = list(differ.compare(content1, content2))
    if not any(line.startswith('- ') or line.startswith('+ ') for line in diff):
        return True
    else:
        for line in diff:
            logging.warning(line)
        return False

def test_md2md_norm_footnote_filter():
    file_path = _check_file_path("./resources/inputs/test_md2md_footnote.md")
    pathlib.Path("./temp").mkdir(parents=True, exist_ok=True)
    output_path = pathlib.Path(f"./temp/{file_path.name}")
    answer_path = pathlib.Path(f"./resources/outputs/{file_path.name}")
    pandoc_command = [
        'pandoc',
        file_path,
        '-o',
        output_path,
        '-f',
        'markdown',
        '-t',
        'gfm',
        '-s',
        '--filter',
        'md2md-norm-footnote-filter'
    ]
    assert subprocess.run(pandoc_command, check=True).returncode == 0
    assert _check_the_same_content(output_path,answer_path)

def test_md2md_norm_footnote_filter_pyio():
    file_path = _check_file_path("./resources/inputs/test_md2md_footnote.md")
    pathlib.Path("./temp").mkdir(parents=True, exist_ok=True)
    output_path = pathlib.Path(f"./temp/{file_path.name}")
    answer_path = pathlib.Path(f"./resources/outputs/{file_path.name}")
    pandoc_filter.run_filters_pyio(file_path,output_path,'markdown','gfm',[pandoc_filter.md2md_norm_footnote_filter])
    assert _check_the_same_content(output_path,answer_path)

def test_md2md_norm_internal_link_filter():
    file_path = _check_file_path("./resources/inputs/test_md2md_internal_link.md")
    pathlib.Path("./temp").mkdir(parents=True, exist_ok=True)
    output_path = pathlib.Path(f"./temp/{file_path.name}")
    answer_path = pathlib.Path(f"./resources/outputs/{file_path.name}")
    pandoc_command = [
        'pandoc',
        file_path,
        '-o',
        output_path,
        '-f',
        'markdown',
        '-t',
        'gfm',
        '-s',
        '--filter',
        'md2md-norm-internal-link-filter'
    ]
    assert subprocess.run(pandoc_command, check=True).returncode == 0
    assert _check_the_same_content(output_path,answer_path)

def test_md2md_norm_internal_link_filter_pyio():
    file_path = _check_file_path("./resources/inputs/test_md2md_internal_link.md")
    pathlib.Path("./temp").mkdir(parents=True, exist_ok=True)
    output_path = pathlib.Path(f"./temp/{file_path.name}")
    answer_path = pathlib.Path(f"./resources/outputs/{file_path.name}")
    pandoc_filter.run_filters_pyio(file_path,output_path,'markdown','gfm',[pandoc_filter.md2md_norm_internal_link_filter])
    assert _check_the_same_content(output_path,answer_path)


def test_md2md_enhance_equation_filter():
    file_path = _check_file_path("./resources/inputs/test_md2md_math.md")
    pathlib.Path("./temp").mkdir(parents=True, exist_ok=True)
    output_path = pathlib.Path(f"./temp/{file_path.name}")
    answer_path = pathlib.Path(f"./resources/outputs/{file_path.name}")
    pandoc_command = [
        'pandoc',
        file_path,
        '-o',
        output_path,
        '-f',
        'markdown',
        '-t',
        'gfm',
        '-s',
        '--filter',
        'md2md-enhance-equation-filter'
    ]
    assert subprocess.run(pandoc_command, check=True).returncode == 0
    assert _check_the_same_content(output_path,answer_path)

def test_md2md_enhance_equation_filter_pyio():
    file_path = _check_file_path("./resources/inputs/test_md2md_math_metadata.md")
    pathlib.Path("./temp").mkdir(parents=True, exist_ok=True)
    output_path = pathlib.Path(f"./temp/{file_path.name}")
    answer_path = pathlib.Path(f"./resources/outputs/{file_path.name}")
    def finalize(doc:pf.Doc,**kwargs):
        runtime_dict:dict = doc.runtime_dict
        if runtime_dict.get('math'):
            doc.metadata['math'] = doc.runtime_dict['math']
    pandoc_filter.run_filters_pyio(file_path,output_path,'markdown','gfm',[pandoc_filter.md2md_enhance_equation_filter],finalize=finalize)
    assert _check_the_same_content(output_path,answer_path)
    
def test_md2md_upload_figure_to_aliyun_filter():
    file_path = _check_file_path("./resources/inputs/test_md2md_figure.md")
    pathlib.Path("./temp").mkdir(parents=True, exist_ok=True)
    output_path = pathlib.Path(f"./temp/{file_path.name}")
    answer_path = pathlib.Path(f"./resources/outputs/{file_path.name}")
    pandoc_filter.run_filters_pyio(
        file_path,output_path,'markdown','gfm',
        [pandoc_filter.md2md_upload_figure_to_aliyun_filter],doc_path=file_path)
    assert _check_the_same_content(output_path,answer_path)


import re
import typeguard
import panflute as pf

from ...utils import TracingLogger,DocRuntimeDict

r"""A pandoc filter that mainly for converting `markdown` to `markdown`.
Enhance math equations.
Specifically, this filter will:
    - Adapt AMS rule for math formula.
        - Auto numbering markdown formulations within \begin{equation} \end{equation}, as in Typora.
    - Allow multiple tags, but only take the first one.
    - Allow multiple labels, but only take the first one.
"""

def _prepare_enhance_equation(doc:pf.Doc):
    doc.runtime_dict = DocRuntimeDict(
        {'equations_count':0,
        'math':False})

@typeguard.typechecked
def _enhance_equation(elem:pf.Element,doc:pf.Doc,**kwargs)->None:
    r"""Follow the general procedure of [Panflute](http://scorreia.com/software/panflute/)
    An action to enhance math equations.
    [modify elements in place]
    
    To make equations recognized correctly, 
        content like the following irregular syntax should to be avoided (unnecessary line breaks):
    ```markdown
    \begin{equation}\tag{33124124}
    e=mc^2 \\
    e=mc^2 \\

    \begin{aligned}
    \\
    \\
    \\
    e=mc^2 \\
    e=mc^2 \\
    \end{aligned}
    \end{equation}
    ```
    """
    typeguard.check_type(kwargs['tracing_logger'],TracingLogger)
    tracing_logger:TracingLogger = kwargs['tracing_logger']
    
    if isinstance(elem, pf.elements.Math):
        doc.runtime_dict['math'] = True
        if elem.format == "DisplayMath":
            tracing_logger.mark(elem)
            text = elem.text
            # delete all labels and tags but record the first one
            first_label = ''
            if matched:= re.search(r"(\\label{[^{}]*})",text): 
                first_label = matched.group(0)
                text = re.sub(r"(\\label{[^{}]*})",'',text)
            first_tag = ''
            if matched:= re.search(r"(\\tag{[^{}]*})",text): 
                first_tag = matched.group(0)
                text = re.sub(r"(\\tag{[^{}]*})",'',text)
            
            if (re.search(r"^\s*\\begin{equation}",text) and re.search(r"\\end{equation}\s*$",text)):
                text = re.sub(r"^\s*\\begin{equation}",'',text)
                text = re.sub(r"\\end{equation}\s*$",'',text)
                
                if first_tag != '':
                    text = f"\\begin{{equation}}{first_label}{first_tag}\n{text.strip(" \n")}\n\\end{{equation}}"
                else:
                    doc.runtime_dict['equations_count'] += 1
                    text = f"\\begin{{equation}}{first_label}\\tag{{{doc.runtime_dict['equations_count']}}}\n{text.strip(" \n")}\n\\end{{equation}}"
            else:
                text = f"{text}\n{first_label}{first_tag}"
            elem.text = f"\n{text.strip(" \n")}\n"
            tracing_logger.check_and_log('equation',elem)

def enhance_equation_filter(doc:pf.Doc=None,**kwargs):
    return pf.run_filters(actions=[_enhance_equation],prepare=_prepare_enhance_equation,doc=doc,tracing_logger=TracingLogger(),**kwargs)


"""
See https://github.com/Zhaopudark/pandoc-filter for documentation.
"""
import logging

from .filters.md2md import enhance_equation_filter as md2md_enhance_equation_filter
from .filters.md2md import norm_footnote_filter as md2md_norm_footnote_filter
from .filters.md2md import norm_internal_link_filter as md2md_norm_internal_link_filter
from .filters.md2md import upload_figure_to_aliyun_filter as md2md_upload_figure_to_aliyun_filter
from .filters.md2html import centralize_figure_filter as md2html_centralize_figure_filter
from .filters.md2html import enhance_link_like_filter as md2html_enhance_link_like_filter
from .filters.md2html import hash_anchor_and_internal_link_filter as md2html_hash_anchor_and_internal_link_filter

from .scripts import run_filters_pyio

from .version import __version__

from .utils import TracingLogger

logger = TracingLogger("./logs/pandoc_filter_log",level=logging.DEBUG)


import typeguard
import panflute as pf

from ...utils import TracingLogger

r"""A pandoc filter that mainly for converting `markdown` to `html`.
Enhance the link-like string to a `link` element.
"""


def _enhance_link_like(elem:pf.Element,doc:pf.Doc,**kwargs)->pf.Link|None:
    r"""Follow the general procedure of [Panflute](http://scorreia.com/software/panflute/)
    An action to process a string that may be like a link.
    Replace the string with a `Link` element.
    [replace elements]
    """
    typeguard.check_type(kwargs['tracing_logger'],TracingLogger)
    tracing_logger:TracingLogger = kwargs['tracing_logger']
    if isinstance(elem, pf.Str) and elem.text.lower().startswith('http'):
        tracing_logger.mark(elem)
        link = pf.Link(elem,url=elem.text)
        tracing_logger.check_and_log('link_like',elem)
        return link

def enhance_link_like_filter(doc:pf.Doc=None,**kwargs)->pf.Doc:
    return pf.run_filters(actions=[_enhance_link_like],doc=doc,tracing_logger=TracingLogger(),**kwargs)




import pathlib
from typing import Callable

import panflute as pf
import typeguard

@typeguard.typechecked
def run_filters_pyio(input_path:pathlib.Path,
                     output_path:pathlib.Path,
                     input_format:str,
                     output_format:str,
                     actions:list[Callable],
                     prepare:Callable|None=None,
                     finalize:Callable|None=None,
                     doc:pf.Doc|None=None,
                     **kwargs):
    with open(input_path, "r", encoding='utf-8') as f:
        markdown_content = f.read()
    doc = pf.convert_text(markdown_content,input_format=input_format,output_format='panflute',standalone=True)
    if prepare:
        prepare(doc=doc,**kwargs)
    for action in actions:
        doc = action(doc=doc,**kwargs)
    if finalize:
        finalize(doc=doc,**kwargs)
    with open(output_path, "w", encoding="utf-8") as f:
        text = pf.convert_text(doc,input_format='panflute',output_format=output_format,standalone=True)
        if not text.endswith('\n'):
            text += '\n'
        f.write(text)



import hashlib
import urllib.parse
import pathlib
import typeguard

def get_text_hash(text:str)->str:
    """
    Normalize the input text to a hash string. This is the global hash function for this project.
    
    There are several steps to normalize the input text:
        1. Decode the url-encoded text.
        2. Strip the leading and trailing spaces.
        3. Lower the text.
        4. Hash the text.
    """
    text = urllib.parse.unquote(text).strip(' ').lower() # 需要统一小写以获取更大的兼容性
    return hashlib.sha256(text.encode('utf-8')).hexdigest()

@typeguard.typechecked
def get_text_file_hash(file_path:pathlib.Path)->str:
    with open(file_path, "r", encoding="utf-8") as file:
        file_data  = file.read()
    return hashlib.sha256(file_data.encode('utf-8')).hexdigest()

@typeguard.typechecked
def get_raw_file_hash(file_path:pathlib.Path)->str:
    with open(file_path, "rb") as file:
        file_data  = file.read()
    return hashlib.sha256(file_data).hexdigest()


import typeguard
import urllib.parse
import panflute as pf

from ...utils import TracingLogger
from ...utils import get_html_href,sub_html_href

r"""A pandoc filter that mainly for converting `markdown` to `markdown`.
Normalize internal links' URLs. Decode the URL if it is URL-encoded.

Anchor:
    An anchor represents a section of a document.
    If in markdown, it is a header or a raw-HTML element with id attribute. Such as:
        headings: `## aaa`
        raw-HTML: `<a id="aaa"></a>`
    If in html, it is a raw-HTML element with id attribute. Such as:
        `<a id="aaa"></a>`
   
Internal Link:
    A internal link is a link that points to an anchor in the same document.
    If in markdown, it is a link with a URL that starts with `#`. Such as:
       md internal links: `[bbb](#aaa)`
       raw-HTML internal links: `<a href="#aaa">bbb</a>`
    If in html, it is a raw-HTML element with href attribute that starts with `#`. Such as:
        `<a href="#aaa">bbb</a>`
"""
@typeguard.typechecked
def _decode_internal_link_url(url:str)->str:
    r"""When converting markdown to any type via pandoc, md internal links' URLs may be automatically URL-encoded before any filter works.
    The encoding is done by default and may not be avoided.
    This function is used to decode the URL.
    """
    decoded_url = urllib.parse.unquote(url.lstrip('#'))
    header_mimic = pf.convert_text(f"# {decoded_url}",input_format='markdown',output_format='gfm',standalone=True)
    return f"#{header_mimic.lstrip('# ')}"

def _norm_internal_link(elem:pf.Element,doc:pf.Doc,**kwargs)->None:
    r"""Follow the general procedure of [Panflute](http://scorreia.com/software/panflute/)
    An action to normalize any internal link's url. Decode if it is URL-encoded.
    [modify elements in place]
    """
    typeguard.check_type(kwargs['tracing_logger'],TracingLogger)
    tracing_logger:TracingLogger = kwargs['tracing_logger']
    
    if isinstance(elem, pf.Link) and elem.url.startswith('#'):
        tracing_logger.mark(elem)       
        elem.url = _decode_internal_link_url(elem.url)
        tracing_logger.check_and_log('anchor_links',elem)
    elif isinstance(elem, pf.RawInline) and elem.format == 'html' and (old_href:=get_html_href(elem.text)) and old_href.startswith('#'):
        tracing_logger.mark(elem)
        elem.text = sub_html_href(elem.text,_decode_internal_link_url(old_href))
        tracing_logger.check_and_log('raw_anchor_links',elem)

def norm_internal_link_filter(doc:pf.Doc=None,**kwargs):
    return pf.run_filters(actions=[_norm_internal_link],doc=doc,tracing_logger=TracingLogger(),**kwargs)



from .enhance_equation import enhance_equation_filter
from .norm_footnote import norm_footnote_filter
from .norm_internal_link import norm_internal_link_filter
from .upload_figure_to_aliyun import upload_figure_to_aliyun_filter



import typeguard
import logging
import panflute as pf

from ...utils import TracingLogger

r"""A pandoc filter that mainly for converting `markdown` to `html`.
Centralize the `figure` element by setting inline styles.

Specifically:
    set `figure` attribute to `text-align:center;`
    set `figure`'s `caption` attribute to `color:#858585;`
 
Deprecated.
    The best way is to use CSS files to define global styles and use them by `--css <css_files>`,
    instead of make an action here.
    So, this method is deprecated.        
    
    When converting markdown to html, Pandoc will generate a `Figure` element for each image,
    and copy the `alt_text` in `![image_url](alt_text)` to the `Caption` part of the `Figure`.
    
    And if in many Blog systems, the `alt_text` in `Image` will be copied again into the `Figure`,
    parallel with the original `Caption` part. Leading to dual captions in html.
    
    A not-so-good solution is to delete/elimiate the `Caption` part of the `Figure` element,
    which relys on the post-processing of the html file.
    
    A better way is to set inline styles on the `Figure` element, which does not rely on
    the post-processing components.
    
    Since in `Panflute`, the `Caption` element has no attributes, it is hard to modify it.
    So, we have to insert a `Div` element into the `Caption` element, and modify the `Div`
    element to show our target `alt_text`.
    
    And, it is better if set a global style `text-align:center` for the whole `Figure` element,
    which will not only influence the `Caption` part, but also the `Image` part.
    
    Even though in some Blog systems, the `Figure` element will be centered automatically,
    it is still a good practice to set the style in the `Figure` element, which will not
    be limited by post-processing components.
    
    For specific, the `Figure` element will be modified as:
        - add `text-align:center;` to the `style` attribute of the `Figure` element
        - new a `Caption` element, which includes a new `Div` element that contains the original `alt_text`
        - add `color:#858585;` to the `style` attribute of the `Div` element
        - replace the original `Caption` element with the new one.
"""

def _centralize_figure(elem:pf.Element,doc:pf.Doc,**kwargs)->None:
    r"""Follow the general procedure of [Panflute](http://scorreia.com/software/panflute/)
    An `action` function to centralize the `Figure` element.
    [modify elements in place]
    """
    logging.warning("""
    The `centralize_figure` filter is deprecated. Please use CSS files to define global styles and use them by `--css <css_files>`.
    See https://github.com/Zhaopudark/pandoc-filter/blob/main/src/pandoc_filter/filters/md2html/centralize_figure.py#L13 for more details.
    """)
    typeguard.check_type(kwargs['tracing_logger'],TracingLogger)
    tracing_logger:TracingLogger = kwargs['tracing_logger']
    if isinstance(elem, pf.Figure):
        tracing_logger.mark(elem)
        for img in elem.content:
            if isinstance(img, pf.Image):
                break
        if 'style' in elem.attributes:
            elem.attributes['style'] += "text-align:center;"
        else:
            elem.attributes = {'style':"text-align:center;"}
        centered_div = pf.Div(*elem.caption.content,attributes={'style':"color:#858585;"})
        elem.caption = pf.Caption(centered_div)
        tracing_logger.check_and_log('figure',elem)

def centralize_figure_filter(doc:pf.Doc=None,**kwargs):
    return pf.run_filters(actions= [_centralize_figure],doc=doc,tracing_logger=TracingLogger()**kwargs)


import difflib
import pathlib
import subprocess
import functools
import pandoc_filter
def _check_file_path(file_path:str)->pathlib.Path:
    file_path:pathlib.Path = pathlib.Path(file_path)
    assert file_path.exists()
    assert file_path.is_file()
    return file_path

def _compare_files(file1_path, file2_path):
    with open(file1_path, 'r', encoding='utf-8') as file1:
        content1 = file1.readlines()

    with open(file2_path, 'r', encoding='utf-8') as file2:
        content2 = file2.readlines()

    differ = difflib.Differ()
    diff = list(differ.compare(content1, content2))
    if not any(line.startswith('- ') or line.startswith('+ ') for line in diff):
        return True
    else:
        return False

def test_md2html_anchor_link_filter():
    file_path = _check_file_path("./resources/inputs/test_md2html_anchor_and_link.md")
    pathlib.Path("./temp").mkdir(parents=True, exist_ok=True)
    output_path = pathlib.Path(f"./temp/{file_path.stem}.html")
    answer_path = pathlib.Path(f"./resources/outputs/{file_path.stem}.html")
    pandoc_command = [
        'pandoc',
        file_path,
        '-o',
        output_path,
        '-f',
        'markdown',
        '-t',
        'html',
        '-s',
        '--filter',
        'md2md-norm-internal-link-filter',
        '--filter',
        'md2html-hash-anchor-and-internal-link-filter',
        '--filter',
        'md2html-enhance-link-like-filter',
    ]
    assert subprocess.run(pandoc_command, check=True).returncode == 0
    assert _compare_files(output_path,answer_path)
    
def test_md2html_anchor_link_filter_pyio():
    file_path = _check_file_path("./resources/inputs/test_md2html_anchor_and_link.md")
    pathlib.Path("./temp").mkdir(parents=True, exist_ok=True)
    output_path = pathlib.Path(f"./temp/{file_path.stem}.html")
    answer_path = pathlib.Path(f"./resources/outputs/{file_path.stem}.html")
    
    pandoc_filter.run_filters_pyio(
        file_path,
        output_path,
        'markdown','html',
        [pandoc_filter.md2md_norm_internal_link_filter,
         pandoc_filter.md2html_hash_anchor_and_internal_link_filter,
         pandoc_filter.md2html_enhance_link_like_filter])
    assert _compare_files(output_path,answer_path)


import subprocess
from packaging import version

def get_pandoc_version()->str:
    try:
        output = subprocess.check_output(['pandoc', '--version'], stderr=subprocess.STDOUT, text=True)
        lines = output.splitlines()
        for line in lines:
            if line.startswith("pandoc "):
                version_str = line.split()[1]
                return version_str
    except subprocess.CalledProcessError as e:
        return None

def check_pandoc_version(required_version:str):
    current_version = version.parse(get_pandoc_version())
    required_version = version.parse(required_version)
    if current_version >= required_version:
        return None
    else:
        raise Exception(f"pandoc version {required_version} or higher is required, but you have {current_version}.")


from typing import Any,TypedDict
import typeguard
import pathlib
from collections import UserDict
import panflute as pf

from .logging_helper import TracingLogger
from .html_helper import sub_html_href
from .oss_helper import OssHelper
        
class InternalLink():
    @typeguard.typechecked
    def __init__(self,elem:pf.Link|pf.RawInline,url:str,guessed_url:str|None) -> None:
        self.elem = elem
        self.url = url
        self.guessed_url = guessed_url
    @typeguard.typechecked
    def sub(self,text:str,tracing_logger:TracingLogger)->None:
        tracing_logger.mark(self.elem)
        if isinstance(self.elem, pf.Link):
            self.elem.url = f"#{text}"
        else: # RawInline
            self.elem.text = sub_html_href(self.elem.text,f"#{text}")
        tracing_logger.check_and_log('internal_link',self.elem)

class DocRuntimeDict(TypedDict):
    anchor_count:dict[str,int]|None
    internal_link_record:list[InternalLink]|None
    equations_count:int|None
    math:bool|None
    doc_path:pathlib.Path|None
    oss_helper:OssHelper|None


import re
import functools

import typeguard
import panflute as pf

from ...utils import TracingLogger,DocRuntimeDict,InternalLink
from ...utils import get_html_id,sub_html_id,get_html_href,get_text_hash
from ..md2md.norm_internal_link import _decode_internal_link_url


r"""A pandoc filter that mainly for converting `markdown` to `html`.
Hash both the anchor's `id` and the internal-link's `url` simultaneously.

Anchor:
    An anchor represents a section of a document.
    If in markdown, it is a header or a raw-HTML element with id attribute. Such as:
        headings: `## aaa`
        raw-HTML: `<a id="aaa"></a>`
    If in html, it is a raw-HTML element with id attribute. Such as:
        `<a id="aaa"></a>`
   
Internal Link:
    A internal link is a link that points to an anchor in the same document.
    If in markdown, it is a link with a URL that starts with `#`. Such as:
       md internal links: `[bbb](#aaa)`
       raw-HTML internal links: `<a href="#aaa">bbb</a>`
    If in html, it is a raw-HTML element with href attribute that starts with `#`. Such as:
        `<a href="#aaa">bbb</a>`

Generally, an anchor's id attribute should be equal to the internal-link's url attribute. But in many converting scenarios, due to some unknown restrictions or normalization mechanisms,
if there are some special characters in the anchor's id attribute or the internal-link's url attribute, their texts may be modified automatically. Then, the above equality may not be satisfied,
thus making the link invalid.

So, to maintain the validity of the link, this filter is designed to normalize both the anchor's id and the internal-link's url simultaneously to their hash,
avoiding auto modification by unknown restrictions. (Hash values are all regular legal characters and will generally be retained.)

Sepcifically, this filter will:
    1. calculate any anchor's id hash value with a unified/global hash function
    2. record and numbering each anchor's id hash value
    3. substitute any anchor's id with its hash value and numbering
        the first anchor will be numbered as `<hash value>-1`, the second anchor (if exists) will be numbered as `<hash value>-2`,...
    4. record and guess any internal-link's url with 2 options:
        For example, if the url is '#aaa-2', then the 2 options are:
           `not number`  - link to the first anchor with original id 'aaa-1' and current id '<hash(aaa-2)>-1'
           `number`      - link to the second anchor with original id 'aaa' and current id '<hash(aaa)>-2'
        record both the two options
    6. after recording and dealing with all anchors, deal with the recorded internal-link:
        if
            the `not number` one matches any recorded anchor, substitute the internal-link's url with the `not number` one
        else if 
            the `number` one matches any recorded anchor, substitute the internal-link's url with the `number` one
        else
            log a warning message and do nothing       
"""

def _prepare_hash_anchor_and_internal_link(doc:pf.Doc):
    doc.runtime_dict = DocRuntimeDict(
        {'anchor_count':{},
         'internal_link_record':[]
         })

def _hash_anchor_id(elem:pf.Element,doc:pf.Doc,**kwargs)->None:
    r"""Follow the general procedure of [Panflute](http://scorreia.com/software/panflute/)
    An `action` function to normalize any anchor's `id` to its hash.
    [modify elements in place]
    """
    typeguard.check_type(kwargs['tracing_logger'],TracingLogger)
    tracing_logger:TracingLogger = kwargs['tracing_logger']

    def _text_hash_count(text:str)->str:
        text_hash = get_text_hash(text)
        if text_hash in doc.runtime_dict['anchor_count']: # 按照text_hash值计数, 重复则加1
            doc.runtime_dict['anchor_count'][text_hash] += 1
        else:
            doc.runtime_dict['anchor_count'][text_hash] = 1
        return text_hash
    if isinstance(elem, pf.Header):
        tracing_logger.mark(elem)
        # 获取header文本内容并剔除#号
        header_text = pf.convert_text(elem,input_format='panflute',output_format='gfm',standalone=True).lstrip('#')
        text_hash = _text_hash_count(header_text)
        elem.identifier = f"{text_hash}-{doc.runtime_dict['anchor_count'][text_hash]}"
        tracing_logger.check_and_log('headings anchor',elem)
    elif isinstance(elem, pf.RawInline) and elem.format == 'html' and (raw_id_text:=get_html_id(elem.text)): # 获取id文本内容但不做任何剔除
        tracing_logger.mark(elem)
        text_hash = _text_hash_count(raw_id_text)
        elem.text = sub_html_id(elem.text,f"{text_hash}-{doc.runtime_dict['anchor_count'][text_hash]}")
        tracing_logger.check_and_log('raw-HTML anchor',elem)


def _internal_link_recorder(elem:pf.Element,doc:pf.Doc,**kwargs)->None:
    r"""Follow the general procedure of [Panflute](http://scorreia.com/software/panflute/)
    A action to pre-normalize and record internal links's `url`.
    [modify nothing]
    """    
    def _url_hash_guess(text:str)->str:
        old_url = text.lstrip('#')
        url = get_text_hash(old_url)
        if match:= re.search(r'-(?P<guessed_index>\d+)$', old_url):
            guessed_index = int(match.groupdict()['guessed_index'])  # 从匹配结果中提取数字部分
            if guessed_index == 0:
                guessed_index = 1
            url_striped = old_url[:match.start()]  # 从匹配结果中提取数字前面的部分
            guessed_url_with_num = get_text_hash(url_striped) + f"-{guessed_index}"
        else:
            guessed_url_with_num = None
        return url,guessed_url_with_num
      
    if isinstance(elem, pf.Link) and elem.url.startswith('#'):
        # Olny md internal links need to be decoded since it will be encoded by pandoc before filter.
        decoded_url = _decode_internal_link_url(elem.url) 
        url,guessed_url_with_num = _url_hash_guess(decoded_url)
        doc.runtime_dict['internal_link_record'].append(InternalLink(elem,url=url,guessed_url=guessed_url_with_num))
    elif isinstance(elem, pf.RawInline) and elem.format == 'html' and (old_href:=get_html_href(elem.text)) and old_href.startswith('#'):
        # raw-HTML internal links will not be encoded by pandoc before filter. So there is no need to decode it.
        url,guessed_url_with_num = _url_hash_guess(old_href)
        doc.runtime_dict['internal_link_record'].append(InternalLink(elem,url=url,guessed_url=guessed_url_with_num))

def _finalize_hash_anchor_and_internal_link(doc:pf.Doc,**kwargs): 
    typeguard.check_type(kwargs['tracing_logger'],TracingLogger)
    tracing_logger:TracingLogger = kwargs['tracing_logger']
    
    id_set = set()
    for k,v in doc.runtime_dict['anchor_count'].items():
        for i in range(1,v+1):
            id_set.add(f"{k}-{i}")
    for internal_link in doc.runtime_dict['internal_link_record']:
        internal_link:InternalLink
        if f"{internal_link.url}-1" in id_set:
            internal_link.sub(f"{internal_link.url}-1",tracing_logger)
        elif internal_link.guessed_url in id_set: # None is not in id_set
            internal_link.sub(f"{internal_link.guessed_url}",tracing_logger)
        else:
            tracing_logger.logger.warning(f"{internal_link.elem}")
            tracing_logger.logger.warning(f"The internal link `{internal_link.url}` is invalid and will not be changed because no target header is found.")

def hash_anchor_and_internal_link_filter(doc:pf.Doc=None,**kwargs)->pf.Doc:
    __finalize_hash_anchor_and_internal_link = functools.partial(_finalize_hash_anchor_and_internal_link,tracing_logger=TracingLogger(),**kwargs)
    return pf.run_filters(
        actions= [_hash_anchor_id,_internal_link_recorder],
        prepare=_prepare_hash_anchor_and_internal_link,
        finalize=__finalize_hash_anchor_and_internal_link,
        doc=doc,tracing_logger=TracingLogger(),**kwargs)


import pathlib
import logging
from .logging_helper import _logger_factory
from .hash_helper import get_text_file_hash,get_raw_file_hash
import oss2
from oss2.credentials import EnvironmentVariableCredentialsProvider
from oss2.models import BucketReferer

class OssHelper:
    def __init__(self,endpoint_name:str,bucket_name:str) -> None:
        self.logger = _logger_factory('logs/oss_log',logging.INFO)
        # self.local_cache_dir = pathlib.Path(local_cache_dir)
        # self.local_relative_root = pathlib.Path(local_relative_root)
        # 从环境变量中获取访问凭证。运行本代码示例之前，请确保已设置环境变量OSS_ACCESS_KEY_ID和OSS_ACCESS_KEY_SECRET。
        self.auth = oss2.ProviderAuth(EnvironmentVariableCredentialsProvider())
        self.endpoint_name = endpoint_name
        self.bucket_name = bucket_name
        self.bucket = oss2.Bucket(self.auth,endpoint=self.endpoint_name,bucket_name=self.bucket_name)
        self.domain = self.bucket.list_bucket_cname().cname[0].domain
        
    def get_hashed_file_name(self,file_path:str):
        file_path:pathlib.Path = pathlib.Path(file_path)
        try:
            with open(file_path,'r', encoding='utf-8') as file:
                file.read()
            file_hash = get_text_file_hash(file_path)
        except UnicodeDecodeError:
            file_hash = get_raw_file_hash(file_path)
        return file_hash+file_path.suffix
    def maybe_upload_file_and_get_src(self,file_path:str)->str:
        r"""Try to upload file to oss. Use hash to indentify all files.
            1. Read and calculate the hash of the input file.
            2. Then get a file name based on the hash.
            3. If the file name has not existed in the bucket, upload the file.
            4. Return the url of the file.
        """
        obj_name = self.get_hashed_file_name(file_path)
        if self.bucket.object_exists(obj_name):
            self.logger.info(f"The object {obj_name} has already existed.")
        else:
            self.bucket.put_object_from_file(obj_name, file_path)
            self.logger.info(f"The object {obj_name} has been uploaded.")
        return f"https://{self.domain}/{obj_name}"


import typeguard
import pathlib
import logging
import logging.handlers
from typing import Any

@typeguard.typechecked
def _logger_factory(name:str,level:int=logging.INFO)->logging.Logger:
    r"""Generate a logger with file and console handler.
    Args:
        name: the name of the logger and the indicator of the log file
        level: the level of the logger
    Returns:
        logger: an instance of logging.Logger that has been configured
    """
    pathlib.Path(name).parent.mkdir(parents=True,exist_ok=True)   
    logger = logging.getLogger(name)
    logger.setLevel(level)
    # check if the console handler exists
    console_handler_exists = any(
        isinstance(handler, logging.StreamHandler)
        for handler in logger.handlers
    )
    file_handler_exists = any(
        isinstance(handler, logging.FileHandler)
        for handler in logger.handlers
    )    
    file_handler = logging.handlers.RotatingFileHandler(
        f"{name}.log",
        maxBytes=2*1024*1024,
        backupCount=5,encoding='utf-8')
    file_handler.setLevel(logging.DEBUG)
    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    file_handler.setFormatter(formatter)
    console_handler.setFormatter(formatter)
    if not file_handler_exists:
        logger.addHandler(file_handler)
    if not console_handler_exists:
        logger.addHandler(console_handler)
    logger.propagate = False
    return logger


class SingletonMeta(type):
    _instances = {}
    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]

class TracingLogger(metaclass=SingletonMeta):
    def __init__(self,name:str="./logs/log",level:int=logging.INFO) -> None: 
        self.logger = _logger_factory(name=name,level=level)
        self.__buf = None
    def info(self,indicator,msg:str):
        self.logger.info(f"==={indicator}:{msg}")
    def mark(self,buf:Any):
        self.__buf = str(buf)
    def check_and_log(self,indicator:str,buf:Any):
        if self.__buf != str(buf):
            self.logger.info(f"---{indicator}:{self.__buf}")
            self.logger.info(f"+++{indicator}:{str(buf)}")


import typeguard
import panflute as pf

from ...utils import TracingLogger

r"""A pandoc filter that mainly for converting `markdown` to `markdown`.
Normalize the footnotes. Remove unnecessary `\n` in the footnote content.
"""


def _norm_footnote(elem:pf.Element,doc:pf.Doc,**kwargs)->pf.Note|None:
    r"""Follow the general procedure of [Panflute](http://scorreia.com/software/panflute/)
    An action to process footnotes.
    Remove unnecessary `\n` in the footnote content.
    [replace elements]
    """
    typeguard.check_type(kwargs['tracing_logger'],TracingLogger)
    tracing_logger:TracingLogger = kwargs['tracing_logger']
    
    if isinstance(elem, pf.Note):
        tracing_logger.mark(elem)
        elem = pf.Note(pf.Para(pf.Str(pf.stringify(elem.content).strip(" \n"))))
        tracing_logger.check_and_log('footnote',elem)
        return elem

def norm_footnote_filter(doc:pf.Doc=None,**kwargs):
    return pf.run_filters(actions=[_norm_footnote],doc=doc,tracing_logger=TracingLogger(),**kwargs)


import os
import functools
import pathlib

import typeguard
import panflute as pf

from ...utils import TracingLogger,OssHelper,DocRuntimeDict
from ...utils import get_html_src,sub_html_src

r"""A pandoc filter that mainly for converting `markdown` to `markdown`.
Auto upload local pictures to Aliyun OSS. Replace the original `src` with the new one.

NOTE:
    The following environment variables should be given in advance:
        $Env:OSS_ENDPOINT_NAME
        $Env:OSS_BUCKET_NAME
        $Env:OSS_ACCESS_KEY_ID
        $Env:OSS_ACCESS_KEY_SECRET
    The doc_path should be given in advance.
"""

def _upload_figure_to_aliyun(elem:pf.Element,doc:pf.Doc,**kwargs)->None:
    r"""Follow the general procedure of [Panflute](http://scorreia.com/software/panflute/)
    An `action` function to upload local pictures to Aliyun OSS. Replace the original src with the new one.
    [modify elements in place]
    """
    typeguard.check_type(kwargs['tracing_logger'],TracingLogger)
    tracing_logger:TracingLogger = kwargs['tracing_logger']
    typeguard.check_type(kwargs['oss_helper'],OssHelper)
    oss_helper: OssHelper = kwargs['oss_helper']
    typeguard.check_type(kwargs['doc_path'],pathlib.Path)
    doc_path: pathlib.Path = kwargs['doc_path']
    if isinstance(elem, pf.Image) and (old_src:=str(elem.url)).startswith('.'): # reletive path
        new_src = oss_helper.maybe_upload_file_and_get_src(doc_path.parent/old_src)
        tracing_logger.mark(elem)
        elem.url = new_src
        tracing_logger.check_and_log('image',elem)
    elif isinstance(elem, pf.RawInline) and elem.format == 'html' and (old_src:=get_html_src(elem.text)) and old_src.startswith('.'): # reletive path
            new_src = oss_helper.maybe_upload_file_and_get_src(doc_path.parent/old_src)
            tracing_logger.mark(elem)
            elem.text = sub_html_src(elem.text,new_src)
            tracing_logger.check_and_log('raw_html_img',elem)

@typeguard.typechecked
def upload_figure_to_aliyun_filter(doc:pf.Doc=None,doc_path:pathlib.Path=None,**kwargs):
    assert doc_path.exists(),f"doc_path: {doc_path} does not exist."
    assert os.environ['OSS_ENDPOINT_NAME'], "OSS_ENDPOINT_NAME is not given in environment variables."
    assert os.environ['OSS_BUCKET_NAME'], "OSS_BUCKET_NAME is not given in environment variables."
    assert os.environ['OSS_ACCESS_KEY_ID'], "OSS_ACCESS_KEY_ID is not given in environment variables."
    assert os.environ['OSS_ACCESS_KEY_SECRET'], "OSS_ACCESS_KEY_SECRET is not given in environment variables."
    return pf.run_filters(
        actions=[_upload_figure_to_aliyun],
        doc=doc,
        tracing_logger=TracingLogger(),
        doc_path=doc_path,
        oss_helper=OssHelper(os.environ['OSS_ENDPOINT_NAME'],os.environ['OSS_BUCKET_NAME']),**kwargs)


import re
import typeguard

def __get_attribute(text:str,attribute:str)->str|None:
    match = re.search(rf"{attribute}=['\"]([^'\"]+)['\"]",text)
    if match:
        return str(match.group(1))
    else:
        return None
def __sub_attribute(text:str,attribute:str,sub_string:str)->str:
    return re.sub(rf"({attribute}=['\"])([^'\"]+)(['\"])", f"\\g<1>{sub_string}\\g<3>", text)

# href
@typeguard.typechecked
def get_html_href(text:str)->str|None:
    return __get_attribute(text,"href")
    
@typeguard.typechecked
def sub_html_href(text:str,sub_string:str)->str:
    return __sub_attribute(text,"href",sub_string)

# id
@typeguard.typechecked
def get_html_id(text:str)->str|None:
    return __get_attribute(text,"id")
@typeguard.typechecked
def sub_html_id(text:str,sub_string:str)->str:
    return __sub_attribute(text,"id",sub_string)

# src
@typeguard.typechecked
def get_html_src(text:str)->str|None:
    return __get_attribute(text,"src")
@typeguard.typechecked
def sub_html_src(text:str,sub_string:str)->str:
    return __sub_attribute(text,"src",sub_string)